% Vehicle parameters and constants
% 20180816 allowed for terrain elevations by varying phi 
% Input is vplan and gplan, vplan_po and gplan_po created by prepare.m or GPSproc.m
% (C) 2018, Olaf Czogalla, ifak

clear;
close all;
fileID = fopen("SOC.txt",'w');
write_statevariables=1;

% In following is modelled the Solaris Urbino 12 Electric

% Common parameters
rho       = 1.29;       % Air density under standard conditions 20C 1013.25 mBar 
%rho       = 1.292;     % Air density under norm conditions 0C 1013.25 mBar
C_rr      = 0.01;       % Rolling resistance coefficient tyre on asphalt 
%C_rr      = 0.015;     % Rolling resistance coefficient tyre on concrete 
%C_rr      = 0.02;      % Rolling resistance coefficient tyre on gravel
dt        = 1;          % Simulation step 1 is equal to 1 s
g         = 9.81;       % Gravity in m/s

% Road conditions
phi       = 0;          % in radians! 0..10 % grade <-> 0..5.7 degrees <-> 0..0.1 rad
s         = 0;          % Distance in m
v         = 0;          % Velocity in m/s
a         = 0;          % acceleration in m/s
v_dc_int  = 0;          % Velocity integral increment  

% Vehicle parameters
A_f       = 8.657;      % Vehicle frontal area in m
C_d       = 0.66;       % Drag coefficient (long cylinder/cubeoid)
r_w       = 0.478;      % Wheel radius in m
m_empty   = 13790;      % Empty bus
m_load    = 5100;       % Load 30 seats / 45 standees a 68 kg
APP       = 0;          % Accelerator pedal position 
BPP       = 0;          % Brake pedal position
m         = m_empty + m_load;
m_i       = 1.04*m ;    % Inertial mass


% Motor parameters
k_i       = 0.0167;     % Motor loss constant 
k_c       = 0.0452;     % Motor loss constant 
k_omega   = 5.0664e-05; % Motor loss constant kgm
C         = 628.2974;   % Motor loss constant from reference motor
G         = 22.6;       % Gear final drive ratio 
MaxRPM    = 10990;      % Maximum motor rotational speed in rpm
MinRPM    = 0;          % Minimum motor rotational speed in rpm
MaxTorque = 973;        % Maximum torque of motor in Nm; 2 * 487 Nm
MaxPower  = 250000;     % in W: 2 x 125 kW
F_b       = 0;          % Braking force
T_loss    = 0;          % Torque losses

% Motor Torque chacteristic curve
rpmsamples  =  [50 1500 2000 3000 5000];
torqsamples =  [500 600  650 400  200];

% Battery parameters
% Battery parameters
Capacity      = 240000;  % Wh = 6 x 40 kWh Packs
Accessoryload = 5000;    % e.g. load for HVAC etc. in W
E             = Capacity;
E_integrated  = 0;       % Charging energy taken from grid
SOC           = 100;     % Initial State of charge in %
V_oc          = 600;     % Voltage open circuit
R_int         = 25/1000; % Internal resistance 25 mOhm in Ohm
Ereg          = 0;       % Accumulated regenerated energy of full day cycle
Ireg          = 0;

% Drive plan in velocity per seconds 
load('vplan.mat');
[vplan_m,vplan_n]=size (vplan);

% Drive plan in grade per seconds 
load('gplan.mat');
[gplan_m,gplan_n]=size (gplan);

% Drive plan pull out in velocity per seconds 
load('vplan_po.mat');
[vplan_po_m,vplan_po_n]=size (vplan_po);

% Drive plan in grade per seconds 
load('gplan_po.mat');
[gplan_po_m,gplan_po_n]=size (gplan_po);


% Arrays for state variables           
%%{
plot_o    = zeros(1,vplan_n);
plot_rpm  = zeros(1,vplan_n);
plot_tmot = zeros(1,vplan_n);
plot_fi   = zeros(1,vplan_n);
plot_ftr  = zeros(1,vplan_n);
plot_a    = zeros(1,vplan_n);
plot_v    = zeros(1,vplan_n);
plot_s    = zeros(1,vplan_n);
plot_i    = zeros(1,vplan_n);
plot_vdc  = zeros(1,vplan_n);
plot_soc  = zeros(1,vplan_n);
plot_y    = zeros(1,vplan_n);
plot_APP  = zeros(1,vplan_n);
%%}

%h=waitbar(0);


% Simulation loop

rpm=500;   % motor speed at start in rpm
y=0;
l=0;

% trip plan 1=towards  -1=backwards 300=charge time in seconds
%           2=pull out -2=pull in
trip_plan=[2 0;    % pull out
           1 0;    %1
          -1 300;  %
           1 0;    %2
          -1 300;  %
           1 0;    %3
          -1 300;  %
           1 0;    %4
          -1 300;  %
           1 0;    %5
          -1 300;  %
           1 0;    %6
          -1 300;  %
           1 0;    %7
          -1 300;  %
           1 0;    %8
          -1 300;  %
           1 0;    %9
          -1 300;  %
           1 0;    %10
          -1 300;  %
           1 0;    %11
          -1 300;  %
           1 0;    %12
          -1 300;  %
           1 0;    %13
          -1 300;  %
           1 0;    %14
          -1 300;  %
           1 0;    %15
          -1 300  
           1 0;    %16
          -1 300;  %
           1 0;    %17
          -1 300;  %
           1 0;    %18
          -1 300   %
          -2 0;    % pull in
          ];

ctotal=cputime;
for cycle=1:size(trip_plan,1)
  cput=cputime;
  trip_step=0;
  if trip_plan(cycle,1)==2
     start = 1; incr  = 1; stop  = vplan_po_n;
  end;
  if trip_plan(cycle,1)==1
     start = 1; incr  = 1; stop  = vplan_n;
  end;
  if trip_plan(cycle,1)==-1
     start = vplan_n; incr  = -1; stop  = 1;
  end;
  if trip_plan(cycle,1)==-2
     start = vplan_po_n; incr  = -1; stop  = 1;
  end;
  
  for l=start:incr:stop
    trip_step=trip_step+1;
    %waitbar(l/vplan_n,h,strcat(num2str(cycle),"/",num2str(size(cycleplan,1)))) ;
    %makes it slow

    % Drive plan to follow
    if trip_plan(cycle,1)== 1 drive_cycle_speed = vplan(1, l)    ; end;
    if trip_plan(cycle,1)==-1 drive_cycle_speed = vplan(1, l)    ; end;
    if trip_plan(cycle,1)== 2 drive_cycle_speed = vplan_po(1, l) ; end;
    if trip_plan(cycle,1)==-2 drive_cycle_speed = vplan_po(1, l) ; end;

    v_dc = drive_cycle_speed - v;
    v_dc_int = v_dc_int + v_dc*dt;

    % Terrain grade to follow
    if trip_plan(cycle,1)== 1 phi = gplan(1, l);    end;
    if trip_plan(cycle,1)==-1 phi = gplan(1, l);    end;
    if trip_plan(cycle,1)== 2 phi = gplan_po(1, l); end;
    if trip_plan(cycle,1)==-2 phi = gplan_po(1, l); end;
    
    % PID controller as driver
    PID = 0.15*v_dc+ 0.01*v_dc_int;   
    btwa=min(PID,1);
    APP =max(btwa,0);        % accelerator pedal position saturated from 0 to 1
    btwb=min(PID,0);
    BPP =abs(max(btwb,-1));  % brake pedal position saturated from 0 to 1
    
    % Rotation speed envelope
    omega =  2*pi*rpm/60;                 % in rad/s
    if omega < 2*pi*500/60 omega = 2*pi*500/60; end;
 
    % according to 18:29 in the video
    PositiveTorque     = APP * min(MaxPower/omega,MaxTorque);
    % PositiveTorque     =  interp1(rpmsamples,torqsamples,rpm,'cubic');
    RegenerativeTorque = min(MaxPower/omega,MaxTorque)/2;
   
    % (PID) Driver brakes
    if (BPP > 0)                
       PositiveTorque     = 0;
       RegenTorque = RegenerativeTorque;
       F_b = BPP* m * 1; % GVW * 1 m/s
    else 
       F_b = 0;
       RegenTorque = 0;
    end; 
                
    % Torque envelope limiter
    MotorNetTorque = PositiveTorque  - RegenTorque;
    T_mot = MotorNetTorque;

    % Driveline
    oldrpm = rpm;
    rpm    = v* G*60/(r_w*2*pi);
    rpm=min(rpm,MaxRPM);
    rpm=max(rpm,MinRPM);
    d_rpm  = rpm - oldrpm; % rotational speed increasing or decreasing

    F_tr= (T_mot - T_loss)*G/r_w - F_b;

    % Required motor power 
    P_loss = k_c*T_mot^2 + k_i*omega + k_omega*omega^3 + C;
    P_mot  = T_mot*omega - P_loss;
    MotorPowerInput = T_mot*omega + P_loss;

    % Battery model
    %MotorPowerInput = F_tr;
    Pel_outbatt     = Accessoryload + MotorPowerInput;
    I               = (V_oc - sqrt(V_oc^2 -4*R_int*Pel_outbatt))/(2*R_int);
    Pel_ideal       = I * V_oc;
    Pel_loss        = I*I*R_int;
    Pel_actual      = Pel_ideal - Pel_loss;
    V_term          = V_oc - I* R_int;
    E               = E - I*V_oc*dt/3600;   % E in Wh  -term: 1 step = 1Ws -> 1/3600 Wh  
    SOC             = 100*(E/Capacity); 
    
    % How much energy is regenerated, when I is negative the battery is charged
    % Battery current must be restricted to 100 A 
    if  (I<-100)  Ireg = -100; else Ireg = I; end;
    if  (I<0)     Ereg = Ereg - Ireg*V_oc*dt/3600; end;
    
    % Vehicle (Glider) model
    F_aero =  C_d * A_f * rho * v^2/2 ;     % Aerodynamic drag in kgm/s = N
    F_rr = m * g * C_rr;                    % Rolling resistance force in N
    F_grade = m * g * sin(phi);             % Slope force in N
    F_i = F_tr - F_aero - F_rr- F_grade ;   % Inertial force in N
    a = F_i/m_i;                            % Acceleration in m/s 
    v = v + a*dt;                           % Vehicle speed m/s
    s = s + v*dt;                           % Vehicle travelled distance m

    % Fill arrays
%%{ 
    if cycle == 4 write_statevariables=1; 
    else          write_statevariables=0; end;
    if write_statevariables 
      plot_o(1,l)=omega;
      plot_rpm(1,l)=rpm;
      plot_tmot(1,l)=T_mot;
      plot_fi(1,l)=F_i;
      plot_ftr(1,l)=F_tr;
      plot_a(1,l)=a;
      plot_v(1,l)=v;
      plot_s(1,l)=s;
      plot_I(1,l)=I;
      plot_vdc(1,l)=v_dc;
      plot_soc(1,l)=SOC;
      plot_y(1,l)=y;
      plot_APP(1,l)=APP;
    end;
%%} 
    
    fprintf (fileID,"%3.2f\n",SOC);

%****************** opportunity charging in terminal stop ********************
    if (trip_plan(cycle,2) > 0) && ( (l==vplan_n) || (l==1)) && (trip_step~=1) 
      % charging with 1/2C - OppCharge 240/2 kW @ 600V => 200A
      Icharge= Capacity / V_oc /2 ;
      for  ll=1:trip_plan(cycle,2)
        E = E + Icharge*V_oc*dt/3600;   % E in Wh  -term: 1 step = 1Ws -> 1/3600 Wh  
        E_integrated= E_integrated+Icharge*V_oc*dt/3600;
        SOC = 100*(E/Capacity); 
        SOC = min(SOC,100);
        fprintf (fileID,"%3.2f\n",SOC);
      end; % for
    end; % if    
  end; % for
  printf('cycle: %d ', cycle);
  printf('CPU time: %f seconds\n', cputime-cput);
end; % for

printf('Total CPU time: %f seconds\n', cputime-ctotal);

% Overnight depot charging at 1/8C 240/8 30 kW  50A
Icharge= Capacity / V_oc /8 ;
while SOC <= 99.5
  E = E + Icharge*V_oc*dt/3600;   % E in Wh  -term: 1 step = 1Ws -> 1/3600 Wh  
  E_integrated= E_integrated+Icharge*V_oc*dt/3600;
  SOC = 100*(E/Capacity); 
  SOC = min(SOC,100);
  fprintf (fileID,"%3.2f\n",SOC);
end; % for


fclose(fileID);
%close(h);

%%{
figure('Position',[10 70 1400 740]);

h(1) = subplot(5,1,1); 
plot(vplan,'k');hold on;
plot(plot_v,'g'); 
xlabel("Time in seconds", "fontsize", 18);
ylabel("v in m/s", "fontsize", 18);
set(gca, "fontsize", 14);
legend('vplan','v');
grid;

%h(2) = subplot(5,1,2); 
%plot(plot_a,'c');hold on;
%plot(plot_vdc); 
%legend('a','v_d_c');
%grid;

h(3) = subplot(5,1,2); 
plot(plot_o,'m');hold on;
plot(plot_tmot,'k');
xlabel("Time in seconds", "fontsize", 18);
ylabel("Torque in Nm", "fontsize", 18);
set(gca, "fontsize", 14);
legend('omega','T_m_o_t');
grid;

h(4) = subplot(5,1,3); 
plot(plot_I,'r');
xlabel("Time in seconds", "fontsize", 18);
ylabel("I in Ampere", "fontsize", 18);
set(gca, "fontsize", 14);
legend('I');
grid;

h(5) = subplot(5,1,4); 
plot(plot_soc,"c");
xlabel("Time in seconds", "fontsize", 18);
ylabel("SOC in %", "fontsize", 18);
set(gca, "fontsize", 14);
legend('SOC');
grid;
%%}